#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <signal.h>
#include <stdio.h>

#define BUFSIZE 1024


static int G_break = 0;

/* Funkce checksum pochazi RFC dokumentu. */
unsigned short checksum(unsigned char *addr, int count)
{
  register long sum = 0;
  
  while( count > 1 )  
  {
    /*  This is the inner loop */
    sum += * (unsigned short *) addr;
    addr+=2;
    count -= 2;
  }  
  /*  Add left-over byte, if any */
  if( count > 0 )
    sum += * (unsigned char *) addr;
  /*  Fold 32-bit sum to 16 bits */
  while (sum>>16 != 0)
    sum = (sum & 0xffff) + (sum >> 16);
  return ~sum;
}

// signal handler
void catchsignal (int sig)
{
  if (sig == SIGINT)
  {
    G_break = 1;
  }
}

int main(int argc, char *argv[])
{
  size_t size;
  hostent *host; 
  icmphdr *icmp, *icmpRecv;
  iphdr *ip;
  int sock, length;
  unsigned int ttl = 255;
  sockaddr_in sendSockAddr, receiveSockAddr;
  char buffer[BUFSIZE];
  fd_set mySet;
  timeval tv;
  char *addrString;
  unsigned short int pid = getpid(), p;
  int c = 0;
  int s = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
  int x;
  double z = 0;
  
  
  /* deklarace promennych pro statistiku */
  double y = 0;
  double max = 0;
  double min = 1000000;
  double suma = 0;
  int counter = 0;
    
  if (argc != 2)
  {
    printf("Syntax:\n\t%s addrress\n",argv[0]);
    return -1;
  }
  /* info o pinganem pocitaci */
  if ((host = gethostbyname(argv[1])) == NULL)
  {
    printf("Wrong address\n");
    return -1;
  }
  /* vytvoreni socketu */ 
 if((sock = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP)) == -1)
 {
   printf("Can't create socket\n");
   return -1;
 }
 
 /* if (s == -1)
  {
    perror("Selhalo vytvoreni socketu.");
    exit(1);
  }
   printf("proslo to, s=%d\n", s);*/
     
  setsockopt(sock, SOL_IP, IP_TTL, (const char *)&ttl, sizeof(ttl));
  /* Zaplneni polozky ICMP hlavicky */
  icmp = (icmphdr *)malloc(sizeof(icmphdr));
  if (icmp == NULL)
  {
    printf("Not enought memory\n");
    return -1;
  }
  icmp->type = ICMP_ECHO;
  icmp->code = 0;
  icmp->un.echo.id = pid;
  /* Priprava sendto */
  sendSockAddr.sin_family = AF_INET;
  sendSockAddr.sin_port = 0;
  memcpy(&(sendSockAddr.sin_addr), host->h_addr, host->h_length);

  // install signal handler for CTRL+C
  signal(SIGINT, catchsignal);

  for (p = 1; p <= 20000; p++)
  {
    if (G_break == 1)
    break;
    
    struct timeval out, in;
    icmp->checksum = 0;
    icmp->un.echo.sequence = p;
    icmp->checksum = checksum((unsigned char *)icmp, sizeof(icmphdr));
    sendto(sock, (char *)icmp, sizeof(icmphdr), 0, (sockaddr *)&sendSockAddr, sizeof(sockaddr));
    gettimeofday(&out, NULL);
    tv.tv_sec = 30;
    tv.tv_usec = 0;
    do
    {
      FD_ZERO(&mySet);
      FD_SET(sock, &mySet);
      if (select(sock + 1, &mySet, NULL, NULL, &tv) < 0)
      {
        printf("Select failed\n");
        break;
      }
      if (FD_ISSET(sock, &mySet))
      {
        size = sizeof(sockaddr_in);
        if ((length = recvfrom(sock, buffer, BUFSIZE, 0, (sockaddr *)&receiveSockAddr, &size)) == -1)
        {
           printf("Receiving data error\n");
        }
        /* dorazil blok dat - IP + ICMP hlavicka. */ 
        gettimeofday(&in, NULL);
        ip = (iphdr *) buffer;
        icmpRecv = (icmphdr *) (buffer + ip->ihl * 4);

        if ((icmpRecv->un.echo.id == pid) && (icmpRecv->type == ICMP_ECHOREPLY) && (icmpRecv->un.echo.sequence == p))
        {
          addrString = strdup(inet_ntoa(receiveSockAddr.sin_addr));
          host = gethostbyaddr(&receiveSockAddr.sin_addr, 4, AF_INET);
          printf("%d bytes from %s (%s)", length, (host == NULL? "?" : host->h_name), addrString);
          printf(": icmp_seq=%d ", icmpRecv->un.echo.sequence);
          y = 1e3*(in.tv_sec - out.tv_sec + 1e-6*(in.tv_usec - out.tv_usec));
          z = z + y;
          max = (y > max)?y:max;
          min = (y < min)?y:min;
          suma+=y;
          printf("ttl=%d, rtt=%.3f ms\n", (int)ip->ttl, 1e3 * (in.tv_sec - out.tv_sec + 1e-6 * (in.tv_usec - out.tv_usec)));
          c++;
          free(addrString);
        }
      }
      else
      {
        printf("- time exceeded -\n");
        break;
      }
    } 
    while (!((icmpRecv->un.echo.id == pid) && (icmpRecv->type == ICMP_ECHOREPLY) && (icmpRecv->un.echo.sequence == p)));
    sleep(1);
  }
  x = 100*(p-c-1)/(p-1);
  close(sock);
  printf("-- ping statistics for %s --\n", argv[1]);
  printf("%d packets transmitted, %d received, %d %% packet loss, total time was %f ms.\n", (p-1), c, x, z);
  printf("Statistics rtt: max/min/avg is %lf/%lf/%lf\n",max,min,suma/c);

  if (icmp != NULL)
  {
    free(icmp);
    icmp = NULL;
  }
  return 0;
}

